/********************************************************************************

 Programmed by SEUNG-JO BAE (3/8/95)

  - Goal: Random Access Write

  - Algorithm: Ravi's

  - Input: Source array & its associated pointer array

  - Output: Target array

  - Version: 1.2
            
     - Assumption: 1) s_tot_nproc = t_nproc
                   2) rank = 1
                   3) Block distribution

  - Note: 1) Set all timer (5/27/96)

          2) define all_local_dense_factor (12/31/96)
	     In PLC2, it is possible local dense factor < 1, but > 0.

	  3) Define Hrange_Seed (1/6/97)

********************************************************************************/

#include "global.h"
#include "main-global.h"
#include "raw_type.h"

#define array_print 'i'

Data_Types     data_type = INT;

Operator_Types op_type   = SUM;

int PLC;

/******************************************************************************
 *                                                                            *
 *  Function : void setup_tsc()
 *                                                                            *
 ******************************************************************************/

void setup_tsc()
{
  int i;

  /*===== Clear target array =====*/

  for (i=0; i<t_L[0]; ++i)
    T0[i] = 0;

  for (i=0; i<s_L[0]; ++i) {
    S0_lidx[i] = (Data_Type) i;
    
    P0_lidx[i] = map_to_local(TARGET, P0[i], 0);
    P0_pnum[i] = map_to_proc(TARGET, P0[i], 0);
  }

}
/******************************************************************************
 *                                                                            *
 *  Function : void setup_tsc_complex()
 *                                                                            *
 ******************************************************************************/

COMPLEX_Data_Type Sc[Lmax], Tc[Lmax];

void setup_tsc_complex()
{
  int i;

  for (i=0; i<s_L[0]; ++i) {
    Sc[i].Real = 1;
    Sc[i].Imag = 1;
  }

  for (i=0; i<t_L[0]; ++i) {
    Tc[i].Real = 0;
    Tc[i].Imag = 0;
  }

}

/******************************************************************************
 *                                                                            *
 *  Function : void main_nlc_osc()                                            *
 *                                                                            *
 ******************************************************************************/


void main_nlc_osc()
{

  int i,j;
  double startwtime, endwtime;

  setup_osc();

  MPI_Barrier(WORLD);
  startwtime = (long) clock(); /*MPI_Wtime();*/

  /***/
  nlc_one_1d();
  /***/

  MPI_Barrier(WORLD);
  endwtime =  (long) clock(); /*MPI_Wtime();*/

  nlc_total       = (endwtime-startwtime) / 1000.; /* millisecond */

  /*------------------------- Print arrays ----------------------------------*/

  if (/*pr_control==0 &&*/ t_N[0] <= PrintMax) {
    print_array1(SOURCE, S0, "", array_print);
    print_array1(POINTER, P0, "", 'i');
  }


  if (/*pr_control==0 &&*/ t_N[0] <= 131072 ) /*PrintMax) */;
 
  if (/*pr_control==0 &&*/ t_N[0] <= PrintMax) 
   print_array1(TARGET, T0, "NLC-OSC", array_print);

  if (pr_control==0 && only_master) 
    printf("\n=======> end of NLC-OSC");


}


/******************************************************************************
 *                                                                            *
 *  Function : void main_nplc_tsc()                                           *
 *                                                                            *
 ******************************************************************************/



void main_nplc_tsc()
{

  int i,j;
  double startwtime, endwtime;

  setup_tsc();
  if (data_type == COMPLEX)
    setup_tsc_complex();


  MPI_Barrier(WORLD);
  startwtime = (long) clock(); /*MPI_Wtime();*/

  if (data_type == COMPLEX)
    nplc_tsc(WORLD,
	     Sc, S0_lidx, s_L_all, s_N_all,
	     P0_lidx, P0_pnum, 
	     Tc, t_L_all,
	     data_type, op_type,
	     K_value, PLC, all_local_dense_factor,
	     comm_tsc_alg);

  else
    nplc_tsc(WORLD,
	     S0, S0_lidx, s_L_all, s_N_all,
	     P0_lidx, P0_pnum, 
	     T0, t_L_all,
	     data_type, op_type,
	     K_value, PLC, all_local_dense_factor,
	     comm_tsc_alg);

/***
  nplc_tsc(my_pnum, K_value, t_tot_nproc, 0,
	      S0, S0_lidx, s_L[0], s_L_table[0], s_L_max[0],
	      P0_lidx, P0_pnum, 
	      T0, t_L[0], t_L_table[0], t_L_max[0],
	      comm_tsc_alg);
***/

  MPI_Barrier(WORLD);
  endwtime = (long) clock(); /*MPI_Wtime();*/

  if (!PLC) {
    nlc_raw_total       = (endwtime-startwtime) / 1000.; /* millisecond */

    if (/*pr_control==0 &&*/ t_N[0] <= PrintMax) {
      if (data_type == COMPLEX)
	print_array1_complex(TARGET, Tc, "NLC-TSC",  array_print);
      else
	print_array1(TARGET, T0, "NLC-TSC",  array_print);
    }

    if (pr_control==0) 
      printf("\n-------> end of NLC-TSC");
  }
  else {
    plc_raw_total       = (endwtime-startwtime) / 1000.; /* millisecond */

    if (/*pr_control==0 &&*/ t_N[0] <= PrintMax) {
      if (data_type == COMPLEX)
	print_array1_complex(TARGET, Tc, "NLC-TSC",  array_print);
      else
	print_array1(TARGET, T0, "NLC-TSC",  array_print);
    }

    if (pr_control==0) 
      printf("\n-------> end of PLC-TSC");
  }

}

/******************************************************************************
 *                                                                            *
 *  Function : void main_nlc()                                                *
 *                                                                            *
 ******************************************************************************/


void main_nlc()  
{
  /*---------------------------- Read Input Data  ----------------------------*/

  one_dim_read_input(pr_control);

  if ( s_active || t_active ) {
    sample_data_1d (S0, P0, T0, hot_degree, hot_fnum, Hot,
		    Hrange, seed, pr_control);

    MPI_Barrier(WORLD);
    clock();

  /*------------------------------- NLC-OSC ----------------------------------*/

    MPI_Barrier(WORLD);
    /***/
    main_nlc_osc();
    /***/

  /*------------------------------- NLC-TSC ----------------------------------*/

    MPI_Barrier(WORLD);
    /***/
    PLC = 0;
    main_nplc_tsc();
    /***/

  /*------------------------------- PLC-TSC ----------------------------------*/

    MPI_Barrier(WORLD);
    /***
    PLC = 1;
    main_nplc_tsc();
    ***/

    if ( fmod(hot_fnum, 1.) == 0. )
      hot_num = (int) hot_fnum;
    else
      hot_num = (int) (t_tot_nproc * hot_fnum);

    switch (pr_control) {
    case 0: nlc_long_output();
            break;
    case 1: 
            
    case 2: 
            
    case 3: nplc_short_output();

    }

  } /* if ( s_active || t_active ) */


}




/******************************************************************************
 *                                                                            *
 *  Function : void main()                                                    *
 *                                                                            *
 ******************************************************************************/



void main(argc, argv)
     int argc;
     char *argv[];
{

  int d;

  MPI_Init(&argc,&argv);

  ORI_WORLD = MPI_COMM_WORLD;
  MPI_Comm_size(ORI_WORLD,&ori_nproc);
  MPI_Comm_rank(ORI_WORLD,&ori_my_pnum);
  MPI_Get_processor_name(processor_name,&namelen);

  only_master = ori_my_pnum == master;

  /*--------------------- Process the command line ---------------------------*/

  /******************************* [Usuage] ************************************ 
                                                                
   0       1        2        3         4      5       6           7        8
  raw  pr_control  seed  hot_degree  hot_fnum K  dense_factor  comm_osc comm_tsc
                                              |___________  option  ...........| 
                                                                
   *****************************************************************************/
    
  if ( argc != 5 && argc != 7 && argc != 9 ) {
    printf("\n\n Improper command line. Exit! \n");
    printf("\n [Usuage] raw pr_control seed degree hot_fnum (K dense_factor (comm_osc comm_tsc))\n\n");
    exit(0);
  };

  pr_control = atoi(argv[1]);	/* pr_control + sample_control */
  sample_control = pr_control / DIV_pr_control;
  pr_control %= DIV_pr_control;

  Hrange_Seed = atof(argv[2]);

  seed   = (int)Hrange_Seed;
  Hrange = fmod(Hrange_Seed, 1.);
  if ( Hrange == 0. )
    Hrange = 1.0;

  hot_degree = atof(argv[3]); 

  /* hot_num = atoi(argv[4]); */

  hot_fnum = atof(argv[4]); 
    
  if ( argc == 7 || argc == 9) {
    K_value = atoi(argv[5]);
    all_local_dense_factor = atof(argv[6]);
  }
  else   {
    K_value = Default_K;
    all_local_dense_factor = Default_dense_factor;
  }

  if ( argc == 9 ) {
    comm_osc_alg = atoi(argv[7]);
    comm_tsc_alg = atoi(argv[8]);
  }
  else   {
    comm_osc_alg = Default_comm_osc_alg;
    comm_tsc_alg = Default_comm_tsc_alg;
  }

  if ( !( comm_osc_alg == 1 || comm_osc_alg == 2 || 
        comm_osc_alg == 3 || comm_osc_alg == 8 || 
        comm_osc_alg == 9 || comm_osc_alg == 10 ))

    comm_osc_alg = Default_comm_osc_alg;


  if ( !( comm_tsc_alg == 1 || comm_tsc_alg == 2 || 
        comm_tsc_alg == 3 || comm_tsc_alg == 9 ))

    comm_tsc_alg = Default_comm_tsc_alg;


  /*-------------------------- Now start RAW --------------------------------*/

  main_nlc();

  MPI_Finalize();

}


